from ..httpClient import HttpClient
from .historicalMarketData_client import HistoricalMarketDataClient
from ...metaApi.models import date
import pytest
import respx
from httpx import Response
market_data_client_api_url = 'https://mt-market-data-client-api-v1.agiliumtrade.agiliumtrade.ai'
http_client = HttpClient()
client = HistoricalMarketDataClient(http_client, 'header.payload.sign')


@pytest.fixture(autouse=True)
async def run_around_tests():
    global http_client
    global client
    http_client = HttpClient()
    client = HistoricalMarketDataClient(http_client, 'header.payload.sign')


class TestHistoricalMarketDataClient:

    @respx.mock
    @pytest.mark.asyncio
    async def test_download_candles(self):
        """Should download historical candles from API."""
        expected = [{
            'symbol': 'AUDNZD',
            'timeframe': '15m',
            'time': '2020-04-07T03:45:00.000Z',
            'brokerTime': '2020-04-07 06:45:00.000',
            'open': 1.03297,
            'high': 1.06309,
            'low': 1.02705,
            'close': 1.043,
            'tickVolume': 1435,
            'spread': 17,
            'volume': 345
        }]

        rsps = respx.get(f'{market_data_client_api_url}/users/current/accounts/accountId/historical-market-data/'
                         'symbols/AUDNZD/timeframes/15m/candles').mock(return_value=Response(200, json=expected))
        candles = await client.get_historical_candles('accountId', 'AUDNZD', '15m', date('2020-04-07T03:45:00.000Z'), 1)
        expected[0]['time'] = date(expected[0]['time'])
        assert rsps.calls[0].request.url == f'{market_data_client_api_url}/users/current/accounts/accountId/' \
                                            'historical-market-data/symbols/AUDNZD/timeframes/15m/candles' \
                                            '?startTime=2020-04-07T03%3A45%3A00.000Z&limit=1'
        assert rsps.calls[0].request.method == 'GET'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert candles == expected

    @respx.mock
    @pytest.mark.asyncio
    async def test_download_ticks(self):
        """Should download historical ticks from API."""
        expected = [{
            'symbol': 'AUDNZD',
            'time': '2020-04-07T03:45:00.000Z',
            'brokerTime': '2020-04-07 06:45:00.000',
            'bid': 1.05297,
            'ask': 1.05309,
            'last': 0.5298,
            'volume': 0.13,
            'side': 'buy'
        }]

        rsps = respx.get(f'{market_data_client_api_url}/users/current/accounts/accountId/historical-market-data/'
                         'symbols/AUDNZD/ticks').mock(return_value=Response(200, json=expected))
        ticks = await client.get_historical_ticks('accountId', 'AUDNZD', date('2020-04-07T03:45:00.000Z'), 0, 1)
        expected[0]['time'] = date(expected[0]['time'])
        assert rsps.calls[0].request.url == f'{market_data_client_api_url}/users/current/accounts/accountId/' \
                                            'historical-market-data/symbols/AUDNZD/ticks' \
                                            '?startTime=2020-04-07T03%3A45%3A00.000Z&offset=0&limit=1'
        assert rsps.calls[0].request.method == 'GET'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert ticks == expected



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

install_requires = [
   'aiohttp==3.7.4', 'python-engineio==3.14.2', 'typing-extensions==3.7.4.3', 'iso8601', 'pytz',
   'python-socketio[asyncio_client]==4.6.0', 'requests==2.24.0', 'websockets==8.1', 'httpx==0.16.1',
   'metaapi-cloud-copyfactory-sdk~=2.1.0', 'metaapi-cloud-metastats-sdk~=1.0.0'
]

tests_require = [
      'pytest', 'pytest-mock', 'pytest-asyncio', 'asynctest', 'mock', 'freezegun==1.0.0', 'respx==0.16.3'
]

setuptools.setup(
    name="metaapi_cloud_sdk",
    version="13.2.1",
    author="Agilium Labs LLC",
    author_email="agiliumtrade@agiliumtrade.ai",
    description="SDK for MetaApi, a professional cloud forex API which includes MetaTrader REST API "
                "and MetaTrader websocket API. Supports both MetaTrader 5 (MT5) and MetaTrader 4 (MT4). CopyFactory"
                "copy trading API included. (https://metaapi.cloud)",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    keywords=['metaapi.cloud', 'MetaTrader', 'MetaTrader 5', 'MetaTrader 4', 'MetaTrader5', 'MetaTrader4', 'MT', 'MT4',
              'MT5', 'forex', 'trading', 'API', 'REST', 'websocket', 'client', 'sdk', 'cloud', 'free', 'copy trading',
              'copytrade', 'copy trade', 'trade copying'],
    url="https://github.com/agiliumtrade-ai/metaapi-python-sdk",
    include_package_data=True,
    package_dir={'metaapi_cloud_sdk': 'lib'},
    packages=['metaapi_cloud_sdk'],
    install_requires=install_requires,
    tests_require=tests_require,
    license='SEE LICENSE IN LICENSE',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Other/Proprietary License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


from ..clients.metaApi.metatraderAccount_client import MetatraderAccountUpdateDto
from .historyStorage import HistoryStorage
from typing import Dict, List
from .expertAdvisor import ExpertAdvisor, NewExpertAdvisorDto
from .models import MetatraderCandle, MetatraderTick
from abc import ABC, abstractmethod
from datetime import datetime


class MetatraderAccountModel(ABC):
    """Defines interface for a MetaTrader account class."""

    @property
    @abstractmethod
    def id(self) -> str:
        """Returns account id.

        Returns:
            Account id.
        """

    @property
    @abstractmethod
    def name(self) -> str:
        """Returns account name.

        Returns:
            Account name.
        """

    @property
    @abstractmethod
    def type(self) -> str:
        """Returns account type. Possible values are cloud and self-hosted.

        Returns:
            Account type.
        """

    @property
    @abstractmethod
    def login(self) -> str:
        """Returns account login.

        Returns:
            Account login.
        """

    @property
    @abstractmethod
    def server(self) -> str:
        """Returns MetaTrader server which hosts the account.

        Returns:
            MetaTrader server which hosts the account.
        """

    @property
    @abstractmethod
    def provisioning_profile_id(self):
        """Returns id of the account's provisioning profile.

        Returns:
            Id of the account's provisioning profile.
        """

    @property
    @abstractmethod
    def application(self):
        """Returns application name to connect the account to. Currently allowed values are MetaApi and AgiliumTrade.

        Returns:
            Application name to connect the account to.
        """

    @property
    @abstractmethod
    def magic(self):
        """Returns MetaTrader magic to place trades using.

        Returns:
            MetaTrader magic to place trades using.
        """

    @property
    @abstractmethod
    def state(self):
        """Returns account deployment state. One of CREATED, DEPLOYING, DEPLOYED, UNDEPLOYING, UNDEPLOYED, DELETING

        Returns:
            Account deployment state.
        """

    @property
    @abstractmethod
    def connection_status(self):
        """Returns terminal & broker connection status, one of CONNECTED, DISCONNECTED, DISCONNECTED_FROM_BROKER

        Returns:
            Terminal & broker connection status.
        """

    @property
    @abstractmethod
    def access_token(self):
        """Returns authorization access token to be used for accessing single account data.
        Intended to be used in browser API.

        Returns:
            Authorization token.
        """

    @property
    @abstractmethod
    def manual_trades(self) -> bool:
        """Returns flag indicating if trades should be placed as manual trades on this account.

        Returns:
            Flag indicating if trades should be placed as manual trades on this account.
        """

    @property
    @abstractmethod
    def extensions(self):
        """Returns API extensions.

        Returns:
            API extensions.
        """

    @property
    @abstractmethod
    def metadata(self) -> Dict:
        """Returns extra information which can be stored together with your account.

        Returns:
            Extra information which can be stored together with your account.
        """

    @property
    @abstractmethod
    def tags(self) -> List[str]:
        """Returns user-defined account tags.

        Returns:
            User-defined account tag.
        """

    @property
    @abstractmethod
    def copy_factory_roles(self) -> List[str]:
        """Returns account roles for CopyFactory2 application.

        Returns:
            Account roles for CopyFactory2 application.
        """

    @property
    @abstractmethod
    def resource_slots(self) -> int:
        """Returns number of resource slots to allocate to account. Allocating extra resource slots
        results in better account performance under load which is useful for some applications. E.g. if you have many
        accounts copying the same strategy via CooyFactory API, then you can increase resourceSlots to get a lower
        trade copying latency. Please note that allocating extra resource slots is a paid option. Default is 1

        Returns:
            Number of resource slots to allocate to account.
        """

    @property
    @abstractmethod
    def base_currency(self) -> str:
        """Returns 3-character ISO currency code of the account base currency. Default value is USD. The setting is to
        be used for copy trading accounts which use national currencies only, such as some Brazilian brokers. You
        should not alter this setting unless you understand what you are doing.

        Returns:
            3-character ISO currency code of the account base currency.
        """

    @property
    @abstractmethod
    def reliability(self) -> str:
        """Returns reliability value. Possible values are regular and high.

        Returns:
            Reliability value.
        """

    @property
    @abstractmethod
    def version(self) -> int:
        """Returns version value. Possible values are 4 and 5.

        Returns:
            Account version value.
        """

    @abstractmethod
    async def reload(self):
        """Reloads MetaTrader account from API.

        Returns:
            A coroutine resolving when MetaTrader account is updated.
        """

    @abstractmethod
    async def remove(self):
        """Removes MetaTrader account. Cloud account transitions to DELETING state.
        It takes some time for an account to be eventually deleted. Self-hosted account is deleted immediately.

        Returns:
            A coroutine resolving when account is scheduled for deletion.
        """

    @abstractmethod
    async def deploy(self):
        """Schedules account for deployment. It takes some time for API server to be started and account to reach the
        DEPLOYED state.

        Returns:
            A coroutine resolving when account is scheduled for deployment.
        """

    @abstractmethod
    async def undeploy(self):
        """Schedules account for undeployment. It takes some time for API server to be stopped and account to reach the
        UNDEPLOYED state.

        Returns:
            A coroutine resolving when account is scheduled for undeployment.
        """

    @abstractmethod
    async def redeploy(self):
        """Schedules account for redeployment. It takes some time for API server to be restarted and account to reach
        the DEPLOYED state.

        Returns:
            A coroutine resolving when account is scheduled for redeployment.
        """

    @abstractmethod
    async def increase_reliability(self):
        """Increases MetaTrader account reliability. The account will be temporary stopped to perform this action.

        Returns:
            A coroutine resolving when account reliability is increased.
        """

    @abstractmethod
    async def wait_deployed(self, timeout_in_seconds=300, interval_in_milliseconds=1000):
        """Waits until API server has finished deployment and account reached the DEPLOYED state.

        Args:
            timeout_in_seconds: Wait timeout in seconds, default is 5m.
            interval_in_milliseconds: Interval between account reloads while waiting for a change, default is 1s.

        Returns:
            A coroutine which resolves when account is deployed.

        Raises:
            TimeoutException: If account has not reached the DEPLOYED state within timeout allowed.
        """

    @abstractmethod
    async def wait_undeployed(self, timeout_in_seconds=300, interval_in_milliseconds=1000):
        """Waits until API server has finished undeployment and account reached the UNDEPLOYED state.

        Args:
            timeout_in_seconds: Wait timeout in seconds, default is 5m.
            interval_in_milliseconds: Interval between account reloads while waiting for a change, default is 1s.

        Returns:
            A coroutine which resolves when account is undeployed.

        Raises:
            TimeoutException: If account have not reached the UNDEPLOYED state within timeout allowed.
        """

    @abstractmethod
    async def wait_removed(self, timeout_in_seconds=300, interval_in_milliseconds=1000):
        """Waits until account has been deleted.

        Args:
            timeout_in_seconds: Wait timeout in seconds, default is 5m.
            interval_in_milliseconds: Interval between account reloads while waiting for a change, default is 1s.

        Returns:
            A coroutine which resolves when account is deleted.

        Raises:
            TimeoutException: If account was not deleted within timeout allowed.
        """

    @abstractmethod
    async def wait_connected(self, timeout_in_seconds=300, interval_in_milliseconds=1000):
        """Waits until API server has connected to the terminal and terminal has connected to the broker.

        Args:
            timeout_in_seconds: Wait timeout in seconds, default is 5m.
            interval_in_milliseconds: Interval between account reloads while waiting for a change, default is 1s.

        Returns:
            A coroutine which resolves when API server is connected to the broker.

        Raises:
            TimeoutException: If account has not connected to the broker within timeout allowed.
        """

    @abstractmethod
    async def connect(self, history_storage: HistoryStorage, history_start_time: datetime = None):
        """Connects to MetaApi.

        Args:
            history_storage: Optional history storage.
            history_start_time: History start time. Used for tests.

        Returns:
            MetaApi connection.
        """

    @abstractmethod
    async def update(self, account: MetatraderAccountUpdateDto):
        """Updates MetaTrader account data.

        Args:
            account: MetaTrader account update.

        Returns:
            A coroutine resolving when account is updated.
        """

    @abstractmethod
    async def get_expert_advisors(self) -> List[ExpertAdvisor]:
        """Retrieves expert advisors of current account.

        Returns:
            A coroutine resolving with an array of expert advisor entities.
        """

    @abstractmethod
    async def get_expert_advisor(self, expert_id: str) -> ExpertAdvisor:
        """Retrieves a expert advisor of current account by id.

        Args:
            expert_id: Expert advisor id.

        Returns:
            A coroutine resolving with expert advisor entity.
        """

    @abstractmethod
    async def create_expert_advisor(self, expert_id: str, expert: NewExpertAdvisorDto) -> ExpertAdvisor:
        """Creates an expert advisor.

        Args:
            expert_id: Expert advisor id.
            expert: Expert advisor data.

        Returns:
            A coroutine resolving with expert advisor entity.
        """

    @abstractmethod
    async def get_historical_candles(self, symbol: str, timeframe: str, start_time: datetime = None,
                                     limit: int = None) -> List[MetatraderCandle]:
        """Returns historical candles for a specific symbol and timeframe from a MetaTrader account.
        See https://metaapi.cloud/docs/client/restApi/api/retrieveMarketData/readHistoricalCandles/

        Args:
            symbol: Symbol to retrieve candles for (e.g. a currency pair or an index).
            timeframe: Defines the timeframe according to which the candles must be generated. Allowed values
            for MT5 are 1m, 2m, 3m, 4m, 5m, 6m, 10m, 12m, 15m, 20m, 30m, 1h, 2h, 3h, 4h, 6h, 8h, 12h, 1d, 1w, 1mn.
            Allowed values for MT4 are 1m, 5m, 15m 30m, 1h, 4h, 1d, 1w, 1mn.
            start_time: Time to start loading candles from. Note that candles are loaded in backwards direction, so
            this should be the latest time. Leave empty to request latest candles.
            limit: Maximum number of candles to retrieve. Must be less or equal to 1000.

        Returns:
            A coroutine resolving with historical candles downloaded.
        """

    @abstractmethod
    async def get_historical_ticks(self, symbol: str, start_time: datetime = None, offset: int = None,
                                   limit: int = None) -> List[MetatraderTick]:
        """Returns historical ticks for a specific symbol from a MetaTrader account.
        See https://metaapi.cloud/docs/client/restApi/api/retrieveMarketData/readHistoricalTicks/

        Args:
            symbol: Symbol to retrieve ticks for (e.g. a currency pair or an index).
            start_time: Time to start loading ticks from. Note that ticks are loaded in backwards direction, so
            this should be the latest time. Leave empty to request latest ticks.
            offset: Number of ticks to skip (you can use it to avoid requesting ticks from previous request twice)
            limit: Maximum number of ticks to retrieve. Must be less or equal to 1000.

        Returns:
            A coroutine resolving with historical ticks downloaded.
        """

    @abstractmethod
    async def _delay(self, timeout_in_milliseconds):
        pass



from .metaApi.metaApi import MetaApi
from .metaApi.historyStorage import HistoryStorage
from .metaApi.memoryHistoryStorage import MemoryHistoryStorage
from .clients.metaApi.synchronizationListener import SynchronizationListener
from .metaApi.models import format_error, format_date, date
from metaapi_cloud_copyfactory_sdk import CopyFactory
from metaapi_cloud_metastats_sdk import MetaStats



from .errorHandler import ValidationException


class OptionsValidator:
    """Class for validating API options."""

    def validate_number(self, value: int or float or None, default_value: int or float, name: str):
        """Validates a number parameter.

        Args:
            value: Value to validate.
            default_value: Default value for an option.
            name: Option name.

        Returns:
            Validated value.

        Raises:
            ValidationException: If value is invalid.
        """
        if value is None:
            return default_value
        if (not isinstance(value, int)) and (not isinstance(value, float)):
            raise ValidationException(f'Parameter {name} must be a number')
        if value < 0:
            raise ValidationException(f'Parameter {name} cannot be lower than 0')
        return value

    def validate_non_zero(self, value: int or float or None, default_value: int or float, name: str):
        """Validates a number parameter to be above zero.

        Args:
            value: Value to validate.
            default_value: Default value for an option.
            name: Option name.

        Returns:
            Validated value.

        Raises:
            ValidationException: If value is invalid.
        """
        value = self.validate_number(value, default_value, name)
        if value == 0:
            raise ValidationException(f'Parameter {name} must be bigger than 0')
        return value

    def validate_boolean(self, value: bool or None, default_value: int or float, name: str):
        """Validates a number parameter.

        Args:
            value: Value to validate.
            default_value: Default value for an option.
            name: Option name.

        Returns:
            Validated value.

        Raises:
            ValidationException: If value is invalid.
        """
        if value is None:
            return default_value
        if not isinstance(value, bool):
            raise ValidationException(f'Parameter {name} must be a boolean')
        return value



from .optionsValidator import OptionsValidator
import pytest
validator = OptionsValidator()


class TestValidateNumber:
    @pytest.mark.asyncio
    def test_validate_option(self):
        """Should validate option."""
        value = validator.validate_number(3, 5, 'opt')
        assert value == 3

    @pytest.mark.asyncio
    def test_set_option_to_default_value(self):
        """Should set option to default value if not specified."""
        value = validator.validate_number(None, 5, 'opt')
        assert value == 5

    @pytest.mark.asyncio
    def test_allow_zero_value(self):
        """Should allow zero value."""
        value = validator.validate_number(0, 5, 'opt')
        assert value == 0

    @pytest.mark.asyncio
    def test_throw_error_if_value_is_not_number(self):
        """Should throw error if value is not number."""
        try:
            validator.validate_number('test', 5, 'opt')
            raise Exception('ValidationException expected')
        except Exception as err:
            assert err.__class__.__name__ == 'ValidationException'
            assert err.__str__() == 'Parameter opt must be a number, check error.details for more information'

    @pytest.mark.asyncio
    def test_throw_error_if_value_is_negative(self):
        """Should throw error if value is negative."""
        try:
            validator.validate_number(-3, 5, 'opt')
            raise Exception('ValidationException expected')
        except Exception as err:
            assert err.__class__.__name__ == 'ValidationException'
            assert err.__str__() == 'Parameter opt cannot be lower than 0, check error.details for more information'


class TestValidateNonZero:
    @pytest.mark.asyncio
    def test_validate_option(self):
        """Should validate option."""
        value = validator.validate_non_zero(3, 5, 'opt')
        assert value == 3

    @pytest.mark.asyncio
    def test_set_option_to_default_value(self):
        """Should set option to default value if not specified."""
        value = validator.validate_non_zero(None, 5, 'opt')
        assert value == 5

    @pytest.mark.asyncio
    def test_throw_error_if_value_is_zero(self):
        """Should throw error if value is zero."""
        try:
            validator.validate_non_zero(0, 5, 'opt')
            raise Exception('ValidationException expected')
        except Exception as err:
            assert err.__class__.__name__ == 'ValidationException'
            assert err.__str__() == 'Parameter opt must be bigger than 0, check error.details for more information'


class TestValidateBoolean:
    @pytest.mark.asyncio
    def test_validate_option(self):
        """Should validate option."""
        value = validator.validate_boolean(True, False, 'opt')
        assert value is True

    @pytest.mark.asyncio
    def test_set_option_to_default_value(self):
        """Should set option to default value if not specified."""
        value = validator.validate_boolean(None, False, 'opt')
        assert value is False

    @pytest.mark.asyncio
    def test_throw_error_if_value_is_not_boolean(self):
        """Should throw error if value is not boolean."""
        try:
            validator.validate_boolean('test', 5, 'opt')
            raise Exception('ValidationException expected')
        except Exception as err:
            assert err.__class__.__name__ == 'ValidationException'
            assert err.__str__() == 'Parameter opt must be a boolean, check error.details for more information'
